@@ -22,7 +22,6 @@ import (
 
 	"github.com/aws/aws-sdk-go/aws"
 	"github.com/aws/aws-sdk-go/aws/credentials"
-	"github.com/aws/aws-sdk-go/aws/credentials/ec2rolecreds"
 	"github.com/aws/aws-sdk-go/aws/credentials/stscreds"
 	"github.com/aws/aws-sdk-go/aws/ec2metadata"
 	"github.com/aws/aws-sdk-go/aws/session"
@@ -35,22 +34,25 @@ import (
 	config_util "github.com/prometheus/common/config"
 	"github.com/prometheus/prometheus/discovery/targetgroup"
 	"github.com/prometheus/prometheus/util/strutil"
-	yaml_util "github.com/prometheus/prometheus/util/yaml"
 )
 
 const (
-	ec2Label              = model.MetaLabelPrefix + "ec2_"
-	ec2LabelAZ            = ec2Label + "availability_zone"
-	ec2LabelInstanceID    = ec2Label + "instance_id"
-	ec2LabelInstanceState = ec2Label + "instance_state"
-	ec2LabelInstanceType  = ec2Label + "instance_type"
-	ec2LabelPublicDNS     = ec2Label + "public_dns_name"
-	ec2LabelPublicIP      = ec2Label + "public_ip"
-	ec2LabelPrivateIP     = ec2Label + "private_ip"
-	ec2LabelSubnetID      = ec2Label + "subnet_id"
-	ec2LabelTag           = ec2Label + "tag_"
-	ec2LabelVPCID         = ec2Label + "vpc_id"
-	subnetSeparator       = ","
+	ec2Label                = model.MetaLabelPrefix + "ec2_"
+	ec2LabelAZ              = ec2Label + "availability_zone"
+	ec2LabelInstanceID      = ec2Label + "instance_id"
+	ec2LabelInstanceState   = ec2Label + "instance_state"
+	ec2LabelInstanceType    = ec2Label + "instance_type"
+	ec2LabelOwnerID         = ec2Label + "owner_id"
+	ec2LabelPlatform        = ec2Label + "platform"
+	ec2LabelPublicDNS       = ec2Label + "public_dns_name"
+	ec2LabelPublicIP        = ec2Label + "public_ip"
+	ec2LabelPrivateDNS      = ec2Label + "private_dns_name"
+	ec2LabelPrivateIP       = ec2Label + "private_ip"
+	ec2LabelPrimarySubnetID = ec2Label + "primary_subnet_id"
+	ec2LabelSubnetID        = ec2Label + "subnet_id"
+	ec2LabelTag             = ec2Label + "tag_"
+	ec2LabelVPCID           = ec2Label + "vpc_id"
+	subnetSeparator         = ","
 )
 
 var (
@@ -71,8 +73,15 @@ var (
 	}
 )
 
+// Filter is the configuration for filtering EC2 instances.
+type Filter struct {
+	Name   string   `yaml:"name"`
+	Values []string `yaml:"values"`
+}
+
 // SDConfig is the configuration for EC2 based service discovery.
 type SDConfig struct {
+	Endpoint        string             `yaml:"endpoint"`
 	Region          string             `yaml:"region"`
 	AccessKey       string             `yaml:"access_key,omitempty"`
 	SecretKey       config_util.Secret `yaml:"secret_key,omitempty"`
@@ -80,9 +89,7 @@ type SDConfig struct {
 	RoleARN         string             `yaml:"role_arn,omitempty"`
 	RefreshInterval model.Duration     `yaml:"refresh_interval,omitempty"`
 	Port            int                `yaml:"port"`
-
-	// Catches all undefined fields and must be empty after parsing.
-	XXX map[string]interface{} `yaml:",inline"`
+	Filters         []*Filter          `yaml:"filters"`
 }
 
 // UnmarshalYAML implements the yaml.Unmarshaler interface.
@@ -93,9 +100,6 @@ func (c *SDConfig) UnmarshalYAML(unmarshal func(interface{}) error) error {
 	if err != nil {
 		return err
 	}
-	if err := yaml_util.CheckOverflow(c.XXX, "ec2_sd_config"); err != nil {
-		return err
-	}
 	if c.Region == "" {
 		sess, err := session.NewSession()
 		if err != nil {
@@ -108,6 +112,11 @@ func (c *SDConfig) UnmarshalYAML(unmarshal func(interface{}) error) error {
 		}
 		c.Region = region
 	}
+	for _, f := range c.Filters {
+		if len(f.Values) == 0 {
+			return fmt.Errorf("EC2 SD configuration filter values cannot be empty")
+		}
+	}
 	return nil
 }
 
@@ -124,6 +133,7 @@ type Discovery struct {
 	profile  string
 	roleARN  string
 	port     int
+	filters  []*Filter
 	logger   log.Logger
 }
 
@@ -138,11 +148,13 @@ func NewDiscovery(conf *SDConfig, logger log.Logger) *Discovery {
 	}
 	return &Discovery{
 		aws: &aws.Config{
+			Endpoint:    &conf.Endpoint,
 			Region:      &conf.Region,
 			Credentials: creds,
 		},
 		profile:  conf.Profile,
 		roleARN:  conf.RoleARN,
+		filters:  conf.Filters,
 		interval: time.Duration(conf.RefreshInterval),
 		port:     conf.Port,
 		logger:   logger,
@@ -186,16 +198,6 @@ func (d *Discovery) Run(ctx context.Context, ch chan<- []*targetgroup.Group) {
 	}
 }
 
-func (d *Discovery) ec2MetadataAvailable(sess *session.Session) (isAvailable bool) {
-	svc := ec2metadata.New(sess, &aws.Config{
-		MaxRetries: aws.Int(0),
-	})
-
-	isAvailable = svc.Available()
-
-	return isAvailable
-}
-
 func (d *Discovery) refresh() (tg *targetgroup.Group, err error) {
 	t0 := time.Now()
 	defer func() {
@@ -218,17 +220,23 @@ func (d *Discovery) refresh() (tg *targetgroup.Group, err error) {
 		creds := stscreds.NewCredentials(sess, d.roleARN)
 		ec2s = ec2.New(sess, &aws.Config{Credentials: creds})
 	} else {
-		if d.aws.Credentials == nil && d.ec2MetadataAvailable(sess) {
-			creds := ec2rolecreds.NewCredentials(sess)
-			ec2s = ec2.New(sess, &aws.Config{Credentials: creds})
-		} else {
-			ec2s = ec2.New(sess)
-		}
+		ec2s = ec2.New(sess)
 	}
 	tg = &targetgroup.Group{
 		Source: *d.aws.Region,
 	}
-	if err = ec2s.DescribeInstancesPages(nil, func(p *ec2.DescribeInstancesOutput, lastPage bool) bool {
+
+	var filters []*ec2.Filter
+	for _, f := range d.filters {
+		filters = append(filters, &ec2.Filter{
+			Name:   aws.String(f.Name),
+			Values: aws.StringSlice(f.Values),
+		})
+	}
+
+	input := &ec2.DescribeInstancesInput{Filters: filters}
+
+	if err = ec2s.DescribeInstancesPages(input, func(p *ec2.DescribeInstancesOutput, lastPage bool) bool {
 		for _, r := range p.Reservations {
 			for _, inst := range r.Instances {
 				if inst.PrivateIpAddress == nil {
@@ -237,10 +245,22 @@ func (d *Discovery) refresh() (tg *targetgroup.Group, err error) {
 				labels := model.LabelSet{
 					ec2LabelInstanceID: model.LabelValue(*inst.InstanceId),
 				}
+
+				if r.OwnerId != nil {
+					labels[ec2LabelOwnerID] = model.LabelValue(*r.OwnerId)
+				}
+
 				labels[ec2LabelPrivateIP] = model.LabelValue(*inst.PrivateIpAddress)
+				if inst.PrivateDnsName != nil {
+					labels[ec2LabelPrivateDNS] = model.LabelValue(*inst.PrivateDnsName)
+				}
 				addr := net.JoinHostPort(*inst.PrivateIpAddress, fmt.Sprintf("%d", d.port))
 				labels[model.AddressLabel] = model.LabelValue(addr)
 
+				if inst.Platform != nil {
+					labels[ec2LabelPlatform] = model.LabelValue(*inst.Platform)
+				}
+
 				if inst.PublicIpAddress != nil {
 					labels[ec2LabelPublicIP] = model.LabelValue(*inst.PublicIpAddress)
 					labels[ec2LabelPublicDNS] = model.LabelValue(*inst.PublicDnsName)
@@ -252,14 +272,19 @@ func (d *Discovery) refresh() (tg *targetgroup.Group, err error) {
 
 				if inst.VpcId != nil {
 					labels[ec2LabelVPCID] = model.LabelValue(*inst.VpcId)
+					labels[ec2LabelPrimarySubnetID] = model.LabelValue(*inst.SubnetId)
 
+					// Deduplicate VPC Subnet IDs maintaining the order of the network interfaces returned by EC2.
+					var subnets []string
 					subnetsMap := make(map[string]struct{})
 					for _, eni := range inst.NetworkInterfaces {
-						subnetsMap[*eni.SubnetId] = struct{}{}
-					}
-					subnets := []string{}
-					for k := range subnetsMap {
-						subnets = append(subnets, k)
+						if eni.SubnetId == nil {
+							continue
+						}
+						if _, ok := subnetsMap[*eni.SubnetId]; !ok {
+							subnetsMap[*eni.SubnetId] = struct{}{}
+							subnets = append(subnets, *eni.SubnetId)
+						}
 					}
 					labels[ec2LabelSubnetID] = model.LabelValue(
 						subnetSeparator +
