@@ -22,8 +22,9 @@ import (
 	"github.com/prometheus/common/model"
 	"github.com/prometheus/prometheus/discovery/targetgroup"
 	"github.com/prometheus/prometheus/util/strutil"
-	"k8s.io/client-go/pkg/apis/extensions/v1beta1"
+	"k8s.io/api/extensions/v1beta1"
 	"k8s.io/client-go/tools/cache"
+	"k8s.io/client-go/util/workqueue"
 )
 
 // Ingress implements discovery of Kubernetes ingresss.
@@ -31,90 +32,101 @@ type Ingress struct {
 	logger   log.Logger
 	informer cache.SharedInformer
 	store    cache.Store
+	queue    *workqueue.Type
 }
 
 // NewIngress returns a new ingress discovery.
 func NewIngress(l log.Logger, inf cache.SharedInformer) *Ingress {
-	return &Ingress{logger: l, informer: inf, store: inf.GetStore()}
-}
-
-// Run implements the Discoverer interface.
-func (s *Ingress) Run(ctx context.Context, ch chan<- []*targetgroup.Group) {
-	// Send full initial set of pod targets.
-	var initial []*targetgroup.Group
-	for _, o := range s.store.List() {
-		tg := s.buildIngress(o.(*v1beta1.Ingress))
-		initial = append(initial, tg)
-	}
-	select {
-	case <-ctx.Done():
-		return
-	case ch <- initial:
-	}
-
-	// Send target groups for ingress updates.
-	send := func(tg *targetgroup.Group) {
-		select {
-		case <-ctx.Done():
-		case ch <- []*targetgroup.Group{tg}:
-		}
-	}
+	s := &Ingress{logger: l, informer: inf, store: inf.GetStore(), queue: workqueue.NewNamed("ingress")}
 	s.informer.AddEventHandler(cache.ResourceEventHandlerFuncs{
 		AddFunc: func(o interface{}) {
 			eventCount.WithLabelValues("ingress", "add").Inc()
-
-			ingress, err := convertToIngress(o)
-			if err != nil {
-				level.Error(s.logger).Log("msg", "converting to Ingress object failed", "err", err.Error())
-				return
-			}
-			send(s.buildIngress(ingress))
+			s.enqueue(o)
 		},
 		DeleteFunc: func(o interface{}) {
 			eventCount.WithLabelValues("ingress", "delete").Inc()
-
-			ingress, err := convertToIngress(o)
-			if err != nil {
-				level.Error(s.logger).Log("msg", "converting to Ingress object failed", "err", err.Error())
-				return
-			}
-			send(&targetgroup.Group{Source: ingressSource(ingress)})
+			s.enqueue(o)
 		},
 		UpdateFunc: func(_, o interface{}) {
 			eventCount.WithLabelValues("ingress", "update").Inc()
-
-			ingress, err := convertToIngress(o)
-			if err != nil {
-				level.Error(s.logger).Log("msg", "converting to Ingress object failed", "err", err.Error())
-				return
-			}
-			send(s.buildIngress(ingress))
+			s.enqueue(o)
 		},
 	})
+	return s
+}
+
+func (i *Ingress) enqueue(obj interface{}) {
+	key, err := cache.DeletionHandlingMetaNamespaceKeyFunc(obj)
+	if err != nil {
+		return
+	}
+
+	i.queue.Add(key)
+}
+
+// Run implements the Discoverer interface.
+func (i *Ingress) Run(ctx context.Context, ch chan<- []*targetgroup.Group) {
+	defer i.queue.ShutDown()
+
+	if !cache.WaitForCacheSync(ctx.Done(), i.informer.HasSynced) {
+		level.Error(i.logger).Log("msg", "ingress informer unable to sync cache")
+		return
+	}
+
+	go func() {
+		for i.process(ctx, ch) {
+		}
+	}()
 
 	// Block until the target provider is explicitly canceled.
 	<-ctx.Done()
 }
 
+func (i *Ingress) process(ctx context.Context, ch chan<- []*targetgroup.Group) bool {
+	keyObj, quit := i.queue.Get()
+	if quit {
+		return false
+	}
+	defer i.queue.Done(keyObj)
+	key := keyObj.(string)
+
+	namespace, name, err := cache.SplitMetaNamespaceKey(key)
+	if err != nil {
+		return true
+	}
+
+	o, exists, err := i.store.GetByKey(key)
+	if err != nil {
+		return true
+	}
+	if !exists {
+		send(ctx, i.logger, RoleIngress, ch, &targetgroup.Group{Source: ingressSourceFromNamespaceAndName(namespace, name)})
+		return true
+	}
+	eps, err := convertToIngress(o)
+	if err != nil {
+		level.Error(i.logger).Log("msg", "converting to Ingress object failed", "err", err)
+		return true
+	}
+	send(ctx, i.logger, RoleIngress, ch, i.buildIngress(eps))
+	return true
+}
+
 func convertToIngress(o interface{}) (*v1beta1.Ingress, error) {
 	ingress, ok := o.(*v1beta1.Ingress)
 	if ok {
 		return ingress, nil
 	}
 
-	deletedState, ok := o.(cache.DeletedFinalStateUnknown)
-	if !ok {
-		return nil, fmt.Errorf("Received unexpected object: %v", o)
-	}
-	ingress, ok = deletedState.Obj.(*v1beta1.Ingress)
-	if !ok {
-		return nil, fmt.Errorf("DeletedFinalStateUnknown contained non-Ingress object: %v", deletedState.Obj)
-	}
-	return ingress, nil
+	return nil, fmt.Errorf("received unexpected object: %v", o)
 }
 
 func ingressSource(s *v1beta1.Ingress) string {
-	return "ingress/" + s.Namespace + "/" + s.Name
+	return ingressSourceFromNamespaceAndName(s.Namespace, s.Name)
+}
+
+func ingressSourceFromNamespaceAndName(namespace, name string) string {
+	return "ingress/" + namespace + "/" + name
 }
 
 const (
@@ -158,23 +170,32 @@ func pathsFromIngressRule(rv *v1beta1.IngressRuleValue) []string {
 	return paths
 }
 
-func (s *Ingress) buildIngress(ingress *v1beta1.Ingress) *targetgroup.Group {
+func (i *Ingress) buildIngress(ingress *v1beta1.Ingress) *targetgroup.Group {
 	tg := &targetgroup.Group{
 		Source: ingressSource(ingress),
 	}
 	tg.Labels = ingressLabels(ingress)
 
-	schema := "http"
-	if ingress.Spec.TLS != nil {
-		schema = "https"
+	tlsHosts := make(map[string]struct{})
+	for _, tls := range ingress.Spec.TLS {
+		for _, host := range tls.Hosts {
+			tlsHosts[host] = struct{}{}
+		}
 	}
+
 	for _, rule := range ingress.Spec.Rules {
 		paths := pathsFromIngressRule(&rule.IngressRuleValue)
 
+		scheme := "http"
+		_, isTLS := tlsHosts[rule.Host]
+		if isTLS {
+			scheme = "https"
+		}
+
 		for _, path := range paths {
 			tg.Targets = append(tg.Targets, model.LabelSet{
 				model.AddressLabel: lv(rule.Host),
-				ingressSchemeLabel: lv(schema),
+				ingressSchemeLabel: lv(scheme),
 				ingressHostLabel:   lv(rule.Host),
 				ingressPathLabel:   lv(path),
 			})
