@@ -17,16 +17,17 @@ import (
 	"context"
 	"errors"
 	"fmt"
+	"net/http"
 	"time"
 
 	"github.com/go-kit/kit/log"
 	"github.com/gophercloud/gophercloud"
 	"github.com/gophercloud/gophercloud/openstack"
+	"github.com/mwitkow/go-conntrack"
 	"github.com/prometheus/client_golang/prometheus"
 	config_util "github.com/prometheus/common/config"
 	"github.com/prometheus/common/model"
 	"github.com/prometheus/prometheus/discovery/targetgroup"
-	yaml_util "github.com/prometheus/prometheus/util/yaml"
 )
 
 var (
@@ -49,21 +50,23 @@ var (
 
 // SDConfig is the configuration for OpenStack based service discovery.
 type SDConfig struct {
-	IdentityEndpoint string             `yaml:"identity_endpoint"`
-	Username         string             `yaml:"username"`
-	UserID           string             `yaml:"userid"`
-	Password         config_util.Secret `yaml:"password"`
-	ProjectName      string             `yaml:"project_name"`
-	ProjectID        string             `yaml:"project_id"`
-	DomainName       string             `yaml:"domain_name"`
-	DomainID         string             `yaml:"domain_id"`
-	Role             Role               `yaml:"role"`
-	Region           string             `yaml:"region"`
-	RefreshInterval  model.Duration     `yaml:"refresh_interval,omitempty"`
-	Port             int                `yaml:"port"`
-
-	// Catches all undefined fields and must be empty after parsing.
-	XXX map[string]interface{} `yaml:",inline"`
+	IdentityEndpoint            string                `yaml:"identity_endpoint"`
+	Username                    string                `yaml:"username"`
+	UserID                      string                `yaml:"userid"`
+	Password                    config_util.Secret    `yaml:"password"`
+	ProjectName                 string                `yaml:"project_name"`
+	ProjectID                   string                `yaml:"project_id"`
+	DomainName                  string                `yaml:"domain_name"`
+	DomainID                    string                `yaml:"domain_id"`
+	ApplicationCredentialName   string                `yaml:"application_credential_name"`
+	ApplicationCredentialID     string                `yaml:"application_credential_id"`
+	ApplicationCredentialSecret config_util.Secret    `yaml:"application_credential_secret"`
+	Role                        Role                  `yaml:"role"`
+	Region                      string                `yaml:"region"`
+	RefreshInterval             model.Duration        `yaml:"refresh_interval,omitempty"`
+	Port                        int                   `yaml:"port"`
+	AllTenants                  bool                  `yaml:"all_tenants,omitempty"`
+	TLSConfig                   config_util.TLSConfig `yaml:"tls_config,omitempty"`
 }
 
 // OpenStackRole is role of the target in OpenStack.
@@ -88,7 +91,7 @@ func (c *Role) UnmarshalYAML(unmarshal func(interface{}) error) error {
 	case OpenStackRoleHypervisor, OpenStackRoleInstance:
 		return nil
 	default:
-		return fmt.Errorf("Unknown OpenStack SD role %q", *c)
+		return fmt.Errorf("unknown OpenStack SD role %q", *c)
 	}
 }
 
@@ -104,9 +107,9 @@ func (c *SDConfig) UnmarshalYAML(unmarshal func(interface{}) error) error {
 		return fmt.Errorf("role missing (one of: instance, hypervisor)")
 	}
 	if c.Region == "" {
-		return fmt.Errorf("Openstack SD configuration requires a region")
+		return fmt.Errorf("openstack SD configuration requires a region")
 	}
-	return yaml_util.CheckOverflow(c.XXX, "openstack_sd_config")
+	return nil
 }
 
 func init() {
@@ -132,24 +135,46 @@ func NewDiscovery(conf *SDConfig, l log.Logger) (Discovery, error) {
 		}
 	} else {
 		opts = gophercloud.AuthOptions{
-			IdentityEndpoint: conf.IdentityEndpoint,
-			Username:         conf.Username,
-			UserID:           conf.UserID,
-			Password:         string(conf.Password),
-			TenantName:       conf.ProjectName,
-			TenantID:         conf.ProjectID,
-			DomainName:       conf.DomainName,
-			DomainID:         conf.DomainID,
+			IdentityEndpoint:            conf.IdentityEndpoint,
+			Username:                    conf.Username,
+			UserID:                      conf.UserID,
+			Password:                    string(conf.Password),
+			TenantName:                  conf.ProjectName,
+			TenantID:                    conf.ProjectID,
+			DomainName:                  conf.DomainName,
+			DomainID:                    conf.DomainID,
+			ApplicationCredentialID:     conf.ApplicationCredentialID,
+			ApplicationCredentialName:   conf.ApplicationCredentialName,
+			ApplicationCredentialSecret: string(conf.ApplicationCredentialSecret),
 		}
 	}
+	client, err := openstack.NewClient(opts.IdentityEndpoint)
+	if err != nil {
+		return nil, err
+	}
+	tls, err := config_util.NewTLSConfig(&conf.TLSConfig)
+	if err != nil {
+		return nil, err
+	}
+	client.HTTPClient = http.Client{
+		Transport: &http.Transport{
+			IdleConnTimeout: 5 * time.Duration(conf.RefreshInterval),
+			TLSClientConfig: tls,
+			DialContext: conntrack.NewDialContextFunc(
+				conntrack.DialWithTracing(),
+				conntrack.DialWithName("openstack_sd"),
+			),
+		},
+		Timeout: 5 * time.Duration(conf.RefreshInterval),
+	}
 	switch conf.Role {
 	case OpenStackRoleHypervisor:
-		hypervisor := NewHypervisorDiscovery(&opts,
+		hypervisor := NewHypervisorDiscovery(client, &opts,
 			time.Duration(conf.RefreshInterval), conf.Port, conf.Region, l)
 		return hypervisor, nil
 	case OpenStackRoleInstance:
-		instance := NewInstanceDiscovery(&opts,
-			time.Duration(conf.RefreshInterval), conf.Port, conf.Region, l)
+		instance := NewInstanceDiscovery(client, &opts,
+			time.Duration(conf.RefreshInterval), conf.Port, conf.Region, conf.AllTenants, l)
 		return instance, nil
 	default:
 		return nil, errors.New("unknown OpenStack discovery role")
