@@ -22,7 +22,7 @@ import (
 
 // item represents a token or text string returned from the scanner.
 type item struct {
-	typ itemType // The type of this item.
+	typ ItemType // The type of this item.
 	pos Pos      // The starting position, in bytes, of this item in the input string.
 	val string   // The value of this item.
 }
@@ -50,25 +50,25 @@ func (i item) String() string {
 
 // isOperator returns true if the item corresponds to a arithmetic or set operator.
 // Returns false otherwise.
-func (i itemType) isOperator() bool { return i > operatorsStart && i < operatorsEnd }
+func (i ItemType) isOperator() bool { return i > operatorsStart && i < operatorsEnd }
 
 // isAggregator returns true if the item belongs to the aggregator functions.
 // Returns false otherwise
-func (i itemType) isAggregator() bool { return i > aggregatorsStart && i < aggregatorsEnd }
+func (i ItemType) isAggregator() bool { return i > aggregatorsStart && i < aggregatorsEnd }
 
 // isAggregator returns true if the item is an aggregator that takes a parameter.
 // Returns false otherwise
-func (i itemType) isAggregatorWithParam() bool {
+func (i ItemType) isAggregatorWithParam() bool {
 	return i == itemTopK || i == itemBottomK || i == itemCountValues || i == itemQuantile
 }
 
 // isKeyword returns true if the item corresponds to a keyword.
 // Returns false otherwise.
-func (i itemType) isKeyword() bool { return i > keywordsStart && i < keywordsEnd }
+func (i ItemType) isKeyword() bool { return i > keywordsStart && i < keywordsEnd }
 
 // isCompairsonOperator returns true if the item corresponds to a comparison operator.
 // Returns false otherwise.
-func (i itemType) isComparisonOperator() bool {
+func (i ItemType) isComparisonOperator() bool {
 	switch i {
 	case itemEQL, itemNEQ, itemLTE, itemLSS, itemGTE, itemGTR:
 		return true
@@ -78,7 +78,7 @@ func (i itemType) isComparisonOperator() bool {
 }
 
 // isSetOperator returns whether the item corresponds to a set operator.
-func (i itemType) isSetOperator() bool {
+func (i ItemType) isSetOperator() bool {
 	switch i {
 	case itemLAND, itemLOR, itemLUnless:
 		return true
@@ -92,7 +92,7 @@ const LowestPrec = 0 // Non-operators.
 // Precedence returns the operator precedence of the binary
 // operator op. If op is not a binary operator, the result
 // is LowestPrec.
-func (i itemType) precedence() int {
+func (i ItemType) precedence() int {
 	switch i {
 	case itemLOR:
 		return 1
@@ -111,7 +111,7 @@ func (i itemType) precedence() int {
 	}
 }
 
-func (i itemType) isRightAssociative() bool {
+func (i ItemType) isRightAssociative() bool {
 	switch i {
 	case itemPOW:
 		return true
@@ -121,10 +121,10 @@ func (i itemType) isRightAssociative() bool {
 
 }
 
-type itemType int
+type ItemType int
 
 const (
-	itemError itemType = iota // Error occurred, value is error message
+	itemError ItemType = iota // Error occurred, value is error message
 	itemEOF
 	itemComment
 	itemIdentifier
@@ -137,12 +137,14 @@ const (
 	itemRightBracket
 	itemComma
 	itemAssign
+	itemColon
 	itemSemicolon
 	itemString
 	itemNumber
 	itemDuration
 	itemBlank
 	itemTimes
+	itemSpace
 
 	operatorsStart
 	// Operators.
@@ -182,11 +184,6 @@ const (
 
 	keywordsStart
 	// Keywords.
-	itemAlert
-	itemIf
-	itemFor
-	itemLabels
-	itemAnnotations
 	itemOffset
 	itemBy
 	itemWithout
@@ -198,7 +195,7 @@ const (
 	keywordsEnd
 )
 
-var key = map[string]itemType{
+var key = map[string]ItemType{
 	// Operators.
 	"and":    itemLAND,
 	"or":     itemLOR,
@@ -218,11 +215,6 @@ var key = map[string]itemType{
 	"quantile":     itemQuantile,
 
 	// Keywords.
-	"alert":       itemAlert,
-	"if":          itemIf,
-	"for":         itemFor,
-	"labels":      itemLabels,
-	"annotations": itemAnnotations,
 	"offset":      itemOffset,
 	"by":          itemBy,
 	"without":     itemWithout,
@@ -235,7 +227,7 @@ var key = map[string]itemType{
 
 // These are the default string representations for common items. It does not
 // imply that those are the only character sequences that can be lexed to such an item.
-var itemTypeStr = map[itemType]string{
+var itemTypeStr = map[ItemType]string{
 	itemLeftParen:    "(",
 	itemRightParen:   ")",
 	itemLeftBrace:    "{",
@@ -244,9 +236,11 @@ var itemTypeStr = map[itemType]string{
 	itemRightBracket: "]",
 	itemComma:        ",",
 	itemAssign:       "=",
+	itemColon:        ":",
 	itemSemicolon:    ";",
 	itemBlank:        "_",
 	itemTimes:        "x",
+	itemSpace:        "<space>",
 
 	itemSUB:      "-",
 	itemADD:      "+",
@@ -274,7 +268,7 @@ func init() {
 	key["nan"] = itemNumber
 }
 
-func (i itemType) String() string {
+func (i ItemType) String() string {
 	if s, ok := itemTypeStr[i]; ok {
 		return s
 	}
@@ -291,7 +285,7 @@ func (i item) desc() string {
 	return fmt.Sprintf("%s %s", i.typ.desc(), i)
 }
 
-func (i itemType) desc() string {
+func (i ItemType) desc() string {
 	switch i {
 	case itemError:
 		return "error"
@@ -334,6 +328,7 @@ type lexer struct {
 	parenDepth  int  // Nesting depth of ( ) exprs.
 	braceOpen   bool // Whether a { is opened.
 	bracketOpen bool // Whether a [ is opened.
+	gotColon    bool // Whether we got a ':' after [ was opened.
 	stringOpen  rune // Quote rune of the string currently being read.
 
 	// seriesDesc is set when a series description for the testing
@@ -366,7 +361,7 @@ func (l *lexer) backup() {
 }
 
 // emit passes an item back to the client.
-func (l *lexer) emit(t itemType) {
+func (l *lexer) emit(t ItemType) {
 	l.items <- item{t, l.start, l.input[l.start:l.pos]}
 	l.start = l.pos
 }
@@ -442,6 +437,13 @@ func (l *lexer) run() {
 	close(l.items)
 }
 
+// Release resources used by lexer.
+func (l *lexer) close() {
+	for range l.items {
+		// Consume.
+	}
+}
+
 // lineComment is the character that starts a line comment.
 const lineComment = "#"
 
@@ -518,8 +520,15 @@ func lexStatements(l *lexer) stateFn {
 		l.stringOpen = r
 		return lexRawString
 	case isAlpha(r) || r == ':':
-		l.backup()
-		return lexKeywordOrIdentifier
+		if !l.bracketOpen {
+			l.backup()
+			return lexKeywordOrIdentifier
+		}
+		if l.gotColon {
+			return l.errorf("unexpected colon %q", r)
+		}
+		l.emit(itemColon)
+		l.gotColon = true
 	case r == '(':
 		l.emit(itemLeftParen)
 		l.parenDepth++
@@ -539,6 +548,7 @@ func lexStatements(l *lexer) stateFn {
 		if l.bracketOpen {
 			return l.errorf("unexpected left bracket %q", r)
 		}
+		l.gotColon = false
 		l.emit(itemLeftBracket)
 		l.bracketOpen = true
 		return lexDuration
@@ -616,6 +626,7 @@ func lexValueSequence(l *lexer) stateFn {
 	case r == eof:
 		return lexStatements
 	case isSpace(r):
+		l.emit(itemSpace)
 		lexSpace(l)
 	case r == '+':
 		l.emit(itemADD)
