@@ -51,18 +51,6 @@ func (e *ParseErr) Error() string {
 	return fmt.Sprintf("parse error at line %d, char %d: %s", e.Line, e.Pos, e.Err)
 }
 
-// ParseStmts parses the input and returns the resulting statements or any occurring error.
-func ParseStmts(input string) (Statements, error) {
-	p := newParser(input)
-
-	stmts, err := p.parseStmts()
-	if err != nil {
-		return nil, err
-	}
-	err = p.typecheck(stmts)
-	return stmts, err
-}
-
 // ParseExpr returns the expression parsed from the input.
 func ParseExpr(input string) (Expr, error) {
 	p := newParser(input)
@@ -104,14 +92,6 @@ func ParseMetricSelector(input string) (m []*labels.Matcher, err error) {
 	return vs.LabelMatchers, nil
 }
 
-// parseSeriesDesc parses the description of a time series.
-func parseSeriesDesc(input string) (labels.Labels, []sequenceValue, error) {
-	p := newParser(input)
-	p.lex.seriesDesc = true
-
-	return p.parseSeriesDesc()
-}
-
 // newParser returns a new parser.
 func newParser(input string) *parser {
 	p := &parser{
@@ -120,20 +100,6 @@ func newParser(input string) *parser {
 	return p
 }
 
-// parseStmts parses a sequence of statements from the input.
-func (p *parser) parseStmts() (stmts Statements, err error) {
-	defer p.recover(&err)
-	stmts = Statements{}
-
-	for p.peek().typ != itemEOF {
-		if p.peek().typ == itemComment {
-			continue
-		}
-		stmts = append(stmts, p.stmt())
-	}
-	return
-}
-
 // parseExpr parses a single expression from the input.
 func (p *parser) parseExpr() (expr Expr, err error) {
 	defer p.recover(&err)
@@ -167,6 +133,14 @@ func (v sequenceValue) String() string {
 	return fmt.Sprintf("%f", v.value)
 }
 
+// parseSeriesDesc parses the description of a time series.
+func parseSeriesDesc(input string) (labels.Labels, []sequenceValue, error) {
+	p := newParser(input)
+	p.lex.seriesDesc = true
+
+	return p.parseSeriesDesc()
+}
+
 // parseSeriesDesc parses a description of a time series into its metric and value sequence.
 func (p *parser) parseSeriesDesc() (m labels.Labels, vals []sequenceValue, err error) {
 	defer p.recover(&err)
@@ -175,6 +149,9 @@ func (p *parser) parseSeriesDesc() (m labels.Labels, vals []sequenceValue, err e
 
 	const ctx = "series values"
 	for {
+		for p.peek().typ == itemSpace {
+			p.next()
+		}
 		if p.peek().typ == itemEOF {
 			break
 		}
@@ -193,6 +170,11 @@ func (p *parser) parseSeriesDesc() (m labels.Labels, vals []sequenceValue, err e
 			for i := uint64(0); i < times; i++ {
 				vals = append(vals, sequenceValue{omitted: true})
 			}
+			// This is to ensure that there is a space between this and the next number.
+			// This is especially required if the next number is negative.
+			if t := p.expectOneOf(itemSpace, itemEOF, ctx).typ; t == itemEOF {
+				break
+			}
 			continue
 		}
 
@@ -210,19 +192,20 @@ func (p *parser) parseSeriesDesc() (m labels.Labels, vals []sequenceValue, err e
 			p.next()
 			k = math.Float64frombits(value.StaleNaN)
 		} else {
-			p.errorf("expected number or 'stale' in %s but got %s", ctx, t.desc())
+			p.errorf("expected number or 'stale' in %s but got %s (value: %s)", ctx, t.desc(), p.peek())
 		}
 		vals = append(vals, sequenceValue{
 			value: k,
 		})
 
 		// If there are no offset repetitions specified, proceed with the next value.
-		if t := p.peek(); t.typ == itemNumber || t.typ == itemBlank || t.typ == itemIdentifier && t.val == "stale" {
+		if t := p.peek(); t.typ == itemSpace {
+			// This ensures there is a space between every value.
 			continue
 		} else if t.typ == itemEOF {
 			break
 		} else if t.typ != itemADD && t.typ != itemSUB {
-			p.errorf("expected next value or relative expansion in %s but got %s", ctx, t.desc())
+			p.errorf("expected next value or relative expansion in %s but got %s (value: %s)", ctx, t.desc(), p.peek())
 		}
 
 		// Expand the repeated offsets into values.
@@ -244,6 +227,12 @@ func (p *parser) parseSeriesDesc() (m labels.Labels, vals []sequenceValue, err e
 				value: k,
 			})
 		}
+		// This is to ensure that there is a space between this expanding notation
+		// and the next number. This is especially required if the next number
+		// is negative.
+		if t := p.expectOneOf(itemSpace, itemEOF, ctx).typ; t == itemEOF {
+			break
+		}
 	}
 	return m, vals, nil
 }
@@ -314,7 +303,7 @@ func (p *parser) error(err error) {
 }
 
 // expect consumes the next token and guarantees it has the required type.
-func (p *parser) expect(exp itemType, context string) item {
+func (p *parser) expect(exp ItemType, context string) item {
 	token := p.next()
 	if token.typ != exp {
 		p.errorf("unexpected %s in %s, expected %s", token.desc(), context, exp.desc())
@@ -323,7 +312,7 @@ func (p *parser) expect(exp itemType, context string) item {
 }
 
 // expectOneOf consumes the next token and guarantees it has one of the required types.
-func (p *parser) expectOneOf(exp1, exp2 itemType, context string) item {
+func (p *parser) expectOneOf(exp1, exp2 ItemType, context string) item {
 	token := p.next()
 	if token.typ != exp1 && token.typ != exp2 {
 		p.errorf("unexpected %s in %s, expected %s or %s", token.desc(), context, exp1.desc(), exp2.desc())
@@ -348,93 +337,7 @@ func (p *parser) recover(errp *error) {
 			*errp = e.(error)
 		}
 	}
-}
-
-// stmt parses any statement.
-//
-// 		alertStatement | recordStatement
-//
-func (p *parser) stmt() Statement {
-	switch tok := p.peek(); tok.typ {
-	case itemAlert:
-		return p.alertStmt()
-	case itemIdentifier, itemMetricIdentifier:
-		return p.recordStmt()
-	}
-	p.errorf("no valid statement detected")
-	return nil
-}
-
-// alertStmt parses an alert rule.
-//
-//		ALERT name IF expr [FOR duration]
-//			[LABELS label_set]
-//			[ANNOTATIONS label_set]
-//
-func (p *parser) alertStmt() *AlertStmt {
-	const ctx = "alert statement"
-
-	p.expect(itemAlert, ctx)
-	name := p.expect(itemIdentifier, ctx)
-	// Alerts require a Vector typed expression.
-	p.expect(itemIf, ctx)
-	expr := p.expr()
-
-	// Optional for clause.
-	var (
-		duration time.Duration
-		err      error
-	)
-	if p.peek().typ == itemFor {
-		p.next()
-		dur := p.expect(itemDuration, ctx)
-		duration, err = parseDuration(dur.val)
-		if err != nil {
-			p.error(err)
-		}
-	}
-
-	var (
-		lset        labels.Labels
-		annotations labels.Labels
-	)
-	if p.peek().typ == itemLabels {
-		p.expect(itemLabels, ctx)
-		lset = p.labelSet()
-	}
-	if p.peek().typ == itemAnnotations {
-		p.expect(itemAnnotations, ctx)
-		annotations = p.labelSet()
-	}
-
-	return &AlertStmt{
-		Name:        name.val,
-		Expr:        expr,
-		Duration:    duration,
-		Labels:      lset,
-		Annotations: annotations,
-	}
-}
-
-// recordStmt parses a recording rule.
-func (p *parser) recordStmt() *RecordStmt {
-	const ctx = "record statement"
-
-	name := p.expectOneOf(itemIdentifier, itemMetricIdentifier, ctx).val
-
-	var lset labels.Labels
-	if p.peek().typ == itemLeftBrace {
-		lset = p.labelSet()
-	}
-
-	p.expect(itemAssign, ctx)
-	expr := p.expr()
-
-	return &RecordStmt{
-		Name:   name,
-		Labels: lset,
-		Expr:   expr,
-	}
+	p.lex.close()
 }
 
 // expr parses any expression.
@@ -448,6 +351,17 @@ func (p *parser) expr() Expr {
 		// If the next token is not an operator the expression is done.
 		op := p.peek().typ
 		if !op.isOperator() {
+			// Check for subquery.
+			if op == itemLeftBracket {
+				expr = p.subqueryOrRangeSelector(expr, false)
+				if s, ok := expr.(*SubqueryExpr); ok {
+					// Parse optional offset.
+					if p.peek().typ == itemOffset {
+						offset := p.offset()
+						s.Offset = offset
+					}
+				}
+			}
 			return expr
 		}
 		p.next() // Consume operator.
@@ -509,7 +423,7 @@ func (p *parser) expr() Expr {
 	}
 }
 
-func (p *parser) balance(lhs Expr, op itemType, rhs Expr, vecMatching *VectorMatching, returnBool bool) *BinaryExpr {
+func (p *parser) balance(lhs Expr, op ItemType, rhs Expr, vecMatching *VectorMatching, returnBool bool) *BinaryExpr {
 	if lhsBE, ok := lhs.(*BinaryExpr); ok {
 		precd := lhsBE.Op.precedence() - op.precedence()
 		if (precd < 0) || (precd == 0 && op.isRightAssociative()) {
@@ -568,11 +482,7 @@ func (p *parser) unaryExpr() Expr {
 
 	// Expression might be followed by a range selector.
 	if p.peek().typ == itemLeftBracket {
-		vs, ok := e.(*VectorSelector)
-		if !ok {
-			p.errorf("range specification must be preceded by a metric selector, but follows a %T instead", e)
-		}
-		e = p.rangeSelector(vs)
+		e = p.subqueryOrRangeSelector(e, true)
 	}
 
 	// Parse optional offset.
@@ -584,6 +494,8 @@ func (p *parser) unaryExpr() Expr {
 			s.Offset = offset
 		case *MatrixSelector:
 			s.Offset = offset
+		case *SubqueryExpr:
+			s.Offset = offset
 		default:
 			p.errorf("offset modifier must be preceded by an instant or range selector, but follows a %T instead", e)
 		}
@@ -592,13 +504,17 @@ func (p *parser) unaryExpr() Expr {
 	return e
 }
 
-// rangeSelector parses a Matrix (a.k.a. range) selector based on a given
-// Vector selector.
+// subqueryOrRangeSelector parses a Subquery based on given Expr (or)
+// a Matrix (a.k.a. range) selector based on a given Vector selector.
 //
-//		<Vector_selector> '[' <duration> ']'
+//		<Vector_selector> '[' <duration> ']' | <Vector_selector> '[' <duration> ':' [<duration>] ']'
 //
-func (p *parser) rangeSelector(vs *VectorSelector) *MatrixSelector {
-	const ctx = "range selector"
+func (p *parser) subqueryOrRangeSelector(expr Expr, checkRange bool) Expr {
+	ctx := "subquery selector"
+	if checkRange {
+		ctx = "range/subquery selector"
+	}
+
 	p.next()
 
 	var erange time.Duration
@@ -610,14 +526,43 @@ func (p *parser) rangeSelector(vs *VectorSelector) *MatrixSelector {
 		p.error(err)
 	}
 
-	p.expect(itemRightBracket, ctx)
+	var itm item
+	if checkRange {
+		itm = p.expectOneOf(itemRightBracket, itemColon, ctx)
+		if itm.typ == itemRightBracket {
+			// Range selector.
+			vs, ok := expr.(*VectorSelector)
+			if !ok {
+				p.errorf("range specification must be preceded by a metric selector, but follows a %T instead", expr)
+			}
+			return &MatrixSelector{
+				Name:          vs.Name,
+				LabelMatchers: vs.LabelMatchers,
+				Range:         erange,
+			}
+		}
+	} else {
+		itm = p.expect(itemColon, ctx)
+	}
 
-	e := &MatrixSelector{
-		Name:          vs.Name,
-		LabelMatchers: vs.LabelMatchers,
-		Range:         erange,
+	// Subquery.
+	var estep time.Duration
+
+	itm = p.expectOneOf(itemRightBracket, itemDuration, ctx)
+	if itm.typ == itemDuration {
+		estepStr := itm.val
+		estep, err = parseDuration(estepStr)
+		if err != nil {
+			p.error(err)
+		}
+		p.expect(itemRightBracket, ctx)
+	}
+
+	return &SubqueryExpr{
+		Expr:  expr,
+		Range: erange,
+		Step:  estep,
 	}
-	return e
 }
 
 // number parses a number.
@@ -810,7 +755,7 @@ func (p *parser) labelSet() labels.Labels {
 //
 //		'{' [ <labelname> <match_op> <match_string>, ... ] '}'
 //
-func (p *parser) labelMatchers(operators ...itemType) []*labels.Matcher {
+func (p *parser) labelMatchers(operators ...ItemType) []*labels.Matcher {
 	const ctx = "label matching"
 
 	matchers := []*labels.Matcher{}
@@ -994,9 +939,9 @@ func (p *parser) expectType(node Node, want ValueType, context string) {
 // them, but the costs are small and might reveal errors when making changes.
 func (p *parser) checkType(node Node) (typ ValueType) {
 	// For expressions the type is determined by their Type function.
-	// Statements and lists do not have a type but are not invalid either.
+	// Lists do not have a type but are not invalid either.
 	switch n := node.(type) {
-	case Statements, Expressions, Statement:
+	case Expressions:
 		typ = ValueTypeNone
 	case Expr:
 		typ = n.Type()
@@ -1007,25 +952,12 @@ func (p *parser) checkType(node Node) (typ ValueType) {
 	// Recursively check correct typing for child nodes and raise
 	// errors in case of bad typing.
 	switch n := node.(type) {
-	case Statements:
-		for _, s := range n {
-			p.expectType(s, ValueTypeNone, "statement list")
-		}
-	case *AlertStmt:
-		p.expectType(n.Expr, ValueTypeVector, "alert statement")
-
 	case *EvalStmt:
 		ty := p.checkType(n.Expr)
 		if ty == ValueTypeNone {
 			p.errorf("evaluation statement must have a valid expression type but got %s", documentedType(ty))
 		}
 
-	case *RecordStmt:
-		ty := p.checkType(n.Expr)
-		if ty != ValueTypeVector && ty != ValueTypeScalar {
-			p.errorf("record statement must have a valid expression of type instant vector or scalar but got %s", documentedType(ty))
-		}
-
 	case Expressions:
 		for _, e := range n {
 			ty := p.checkType(e)
@@ -1110,6 +1042,12 @@ func (p *parser) checkType(node Node) (typ ValueType) {
 			p.errorf("unary expression only allowed on expressions of type scalar or instant vector, got %q", documentedType(t))
 		}
 
+	case *SubqueryExpr:
+		ty := p.checkType(n.Expr)
+		if ty != ValueTypeVector {
+			p.errorf("subquery is only allowed on instant vector, got %s in %q instead", ty, n.String())
+		}
+
 	case *NumberLiteral, *MatrixSelector, *StringLiteral, *VectorSelector:
 		// Nothing to do for terminals.
 
