@@ -139,7 +139,8 @@ var testExpr = []struct {
 			},
 		},
 	}, {
-		input: "-some_metric", expected: &UnaryExpr{
+		input: "-some_metric",
+		expected: &UnaryExpr{
 			Op: itemSUB,
 			Expr: &VectorSelector{
 				Name: "some_metric",
@@ -149,7 +150,8 @@ var testExpr = []struct {
 			},
 		},
 	}, {
-		input: "+some_metric", expected: &UnaryExpr{
+		input: "+some_metric",
+		expected: &UnaryExpr{
 			Op: itemADD,
 			Expr: &VectorSelector{
 				Name: "some_metric",
@@ -940,10 +942,6 @@ var testExpr = []struct {
 		input:  `foo{__name__="bar"}`,
 		fail:   true,
 		errMsg: "metric name must not be set twice: \"foo\" or \"bar\"",
-		// }, {
-		// 	input:  `:foo`,
-		// 	fail:   true,
-		// 	errMsg: "bla",
 	},
 	// Test matrix selector.
 	{
@@ -1049,11 +1047,11 @@ var testExpr = []struct {
 	}, {
 		input:  `some_metric OFFSET 1m[5m]`,
 		fail:   true,
-		errMsg: "could not parse remaining input \"[5m]\"...",
+		errMsg: "parse error at char 25: unexpected \"]\" in subquery selector, expected \":\"",
 	}, {
 		input:  `(foo + bar)[5m]`,
 		fail:   true,
-		errMsg: "could not parse remaining input \"[5m]\"...",
+		errMsg: "parse error at char 15: unexpected \"]\" in subquery selector, expected \":\"",
 	},
 	// Test aggregation.
 	{
@@ -1339,10 +1337,6 @@ var testExpr = []struct {
 		input:  "e-+=/(0)",
 		fail:   true,
 		errMsg: `no valid expression found`,
-	}, {
-		input:  "-If",
-		fail:   true,
-		errMsg: `no valid expression found`,
 	},
 	// String quoting and escape sequence interpretation tests.
 	{
@@ -1392,13 +1386,207 @@ var testExpr = []struct {
 		fail:   true,
 		errMsg: "illegal character U+002E '.' in escape sequence",
 	},
+	// Subquery.
+	{
+		input: `foo{bar="baz"}[10m:6s]`,
+		expected: &SubqueryExpr{
+			Expr: &VectorSelector{
+				Name: "foo",
+				LabelMatchers: []*labels.Matcher{
+					mustLabelMatcher(labels.MatchEqual, "bar", "baz"),
+					mustLabelMatcher(labels.MatchEqual, string(model.MetricNameLabel), "foo"),
+				},
+			},
+			Range: 10 * time.Minute,
+			Step:  6 * time.Second,
+		},
+	}, {
+		input: `foo[10m:]`,
+		expected: &SubqueryExpr{
+			Expr: &VectorSelector{
+				Name: "foo",
+				LabelMatchers: []*labels.Matcher{
+					mustLabelMatcher(labels.MatchEqual, string(model.MetricNameLabel), "foo"),
+				},
+			},
+			Range: 10 * time.Minute,
+		},
+	}, {
+		input: `min_over_time(rate(foo{bar="baz"}[2s])[5m:5s])`,
+		expected: &Call{
+			Func: mustGetFunction("min_over_time"),
+			Args: Expressions{
+				&SubqueryExpr{
+					Expr: &Call{
+						Func: mustGetFunction("rate"),
+						Args: Expressions{
+							&MatrixSelector{
+								Name:  "foo",
+								Range: 2 * time.Second,
+								LabelMatchers: []*labels.Matcher{
+									mustLabelMatcher(labels.MatchEqual, "bar", "baz"),
+									mustLabelMatcher(labels.MatchEqual, string(model.MetricNameLabel), "foo"),
+								},
+							},
+						},
+					},
+					Range: 5 * time.Minute,
+					Step:  5 * time.Second,
+				},
+			},
+		},
+	}, {
+		input: `min_over_time(rate(foo{bar="baz"}[2s])[5m:])[4m:3s]`,
+		expected: &SubqueryExpr{
+			Expr: &Call{
+				Func: mustGetFunction("min_over_time"),
+				Args: Expressions{
+					&SubqueryExpr{
+						Expr: &Call{
+							Func: mustGetFunction("rate"),
+							Args: Expressions{
+								&MatrixSelector{
+									Name:  "foo",
+									Range: 2 * time.Second,
+									LabelMatchers: []*labels.Matcher{
+										mustLabelMatcher(labels.MatchEqual, "bar", "baz"),
+										mustLabelMatcher(labels.MatchEqual, string(model.MetricNameLabel), "foo"),
+									},
+								},
+							},
+						},
+						Range: 5 * time.Minute,
+					},
+				},
+			},
+			Range: 4 * time.Minute,
+			Step:  3 * time.Second,
+		},
+	}, {
+		input: `min_over_time(rate(foo{bar="baz"}[2s])[5m:] offset 4m)[4m:3s]`,
+		expected: &SubqueryExpr{
+			Expr: &Call{
+				Func: mustGetFunction("min_over_time"),
+				Args: Expressions{
+					&SubqueryExpr{
+						Expr: &Call{
+							Func: mustGetFunction("rate"),
+							Args: Expressions{
+								&MatrixSelector{
+									Name:  "foo",
+									Range: 2 * time.Second,
+									LabelMatchers: []*labels.Matcher{
+										mustLabelMatcher(labels.MatchEqual, "bar", "baz"),
+										mustLabelMatcher(labels.MatchEqual, string(model.MetricNameLabel), "foo"),
+									},
+								},
+							},
+						},
+						Range:  5 * time.Minute,
+						Offset: 4 * time.Minute,
+					},
+				},
+			},
+			Range: 4 * time.Minute,
+			Step:  3 * time.Second,
+		},
+	}, {
+		input: "sum without(and, by, avg, count, alert, annotations)(some_metric) [30m:10s]",
+		expected: &SubqueryExpr{
+			Expr: &AggregateExpr{
+				Op:      itemSum,
+				Without: true,
+				Expr: &VectorSelector{
+					Name: "some_metric",
+					LabelMatchers: []*labels.Matcher{
+						mustLabelMatcher(labels.MatchEqual, string(model.MetricNameLabel), "some_metric"),
+					},
+				},
+				Grouping: []string{"and", "by", "avg", "count", "alert", "annotations"},
+			},
+			Range: 30 * time.Minute,
+			Step:  10 * time.Second,
+		},
+	}, {
+		input: `some_metric OFFSET 1m [10m:5s]`,
+		expected: &SubqueryExpr{
+			Expr: &VectorSelector{
+				Name: "some_metric",
+				LabelMatchers: []*labels.Matcher{
+					mustLabelMatcher(labels.MatchEqual, string(model.MetricNameLabel), "some_metric"),
+				},
+				Offset: 1 * time.Minute,
+			},
+			Range: 10 * time.Minute,
+			Step:  5 * time.Second,
+		},
+	}, {
+		input: `(foo + bar{nm="val"})[5m:]`,
+		expected: &SubqueryExpr{
+			Expr: &ParenExpr{
+				Expr: &BinaryExpr{
+					Op: itemADD,
+					VectorMatching: &VectorMatching{
+						Card: CardOneToOne,
+					},
+					LHS: &VectorSelector{
+						Name: "foo",
+						LabelMatchers: []*labels.Matcher{
+							mustLabelMatcher(labels.MatchEqual, string(model.MetricNameLabel), "foo"),
+						},
+					},
+					RHS: &VectorSelector{
+						Name: "bar",
+						LabelMatchers: []*labels.Matcher{
+							mustLabelMatcher(labels.MatchEqual, "nm", "val"),
+							mustLabelMatcher(labels.MatchEqual, string(model.MetricNameLabel), "bar"),
+						},
+					},
+				},
+			},
+			Range: 5 * time.Minute,
+		},
+	}, {
+		input: `(foo + bar{nm="val"})[5m:] offset 10m`,
+		expected: &SubqueryExpr{
+			Expr: &ParenExpr{
+				Expr: &BinaryExpr{
+					Op: itemADD,
+					VectorMatching: &VectorMatching{
+						Card: CardOneToOne,
+					},
+					LHS: &VectorSelector{
+						Name: "foo",
+						LabelMatchers: []*labels.Matcher{
+							mustLabelMatcher(labels.MatchEqual, string(model.MetricNameLabel), "foo"),
+						},
+					},
+					RHS: &VectorSelector{
+						Name: "bar",
+						LabelMatchers: []*labels.Matcher{
+							mustLabelMatcher(labels.MatchEqual, "nm", "val"),
+							mustLabelMatcher(labels.MatchEqual, string(model.MetricNameLabel), "bar"),
+						},
+					},
+				},
+			},
+			Range:  5 * time.Minute,
+			Offset: 10 * time.Minute,
+		},
+	}, {
+		input:  "test[5d] OFFSET 10s [10m:5s]",
+		fail:   true,
+		errMsg: "parse error at char 29: subquery is only allowed on instant vector, got matrix in \"test[5d] offset 10s[10m:5s]\"",
+	}, {
+		input:  `(foo + bar{nm="val"})[5m:][10m:5s]`,
+		fail:   true,
+		errMsg: "parse error at char 27: could not parse remaining input \"[10m:5s]\"...",
+	},
 }
 
 func TestParseExpressions(t *testing.T) {
 	for _, test := range testExpr {
-		parser := newParser(test.input)
-
-		expr, err := parser.parseExpr()
+		expr, err := ParseExpr(test.input)
 
 		// Unexpected errors are always caused by a bug.
 		if err == errUnexpected {
@@ -1409,6 +1597,7 @@ func TestParseExpressions(t *testing.T) {
 			t.Errorf("error in input '%s'", test.input)
 			t.Fatalf("could not parse: %s", err)
 		}
+
 		if test.fail && err != nil {
 			if !strings.Contains(err.Error(), test.errMsg) {
 				t.Errorf("unexpected error on input '%s'", test.input)
@@ -1417,24 +1606,6 @@ func TestParseExpressions(t *testing.T) {
 			continue
 		}
 
-		err = parser.typecheck(expr)
-		if !test.fail && err != nil {
-			t.Errorf("error on input '%s'", test.input)
-			t.Fatalf("typecheck failed: %s", err)
-		}
-
-		if test.fail {
-			if err != nil {
-				if !strings.Contains(err.Error(), test.errMsg) {
-					t.Errorf("unexpected error on input '%s'", test.input)
-					t.Fatalf("expected error to contain %q but got %q", test.errMsg, err)
-				}
-				continue
-			}
-			t.Errorf("error on input '%s'", test.input)
-			t.Fatalf("failure expected, but passed with result: %q", expr)
-		}
-
 		if !reflect.DeepEqual(expr, test.expected) {
 			t.Errorf("error on input '%s'", test.input)
 			t.Fatalf("no match\n\nexpected:\n%s\ngot: \n%s\n", Tree(test.expected), Tree(expr))
@@ -1444,12 +1615,10 @@ func TestParseExpressions(t *testing.T) {
 
 // NaN has no equality. Thus, we need a separate test for it.
 func TestNaNExpression(t *testing.T) {
-	parser := newParser("NaN")
-
-	expr, err := parser.parseExpr()
+	expr, err := ParseExpr("NaN")
 	if err != nil {
 		t.Errorf("error on input 'NaN'")
-		t.Fatalf("coud not parse: %s", err)
+		t.Fatalf("could not parse: %s", err)
 	}
 
 	nl, ok := expr.(*NumberLiteral)
@@ -1464,257 +1633,6 @@ func TestNaNExpression(t *testing.T) {
 	}
 }
 
-var testStatement = []struct {
-	input    string
-	expected Statements
-	fail     bool
-}{
-	{
-		// Test a file-like input.
-		input: `
-			# A simple test recording rule.
-			dc:http_request:rate5m = sum(rate(http_request_count[5m])) by (dc)
-
-			# A simple test alerting rule.
-			ALERT GlobalRequestRateLow IF(dc:http_request:rate5m < 10000) FOR 5m
-			  LABELS {
-			    service = "testservice"
-			    # ... more fields here ...
-			  }
-			  ANNOTATIONS {
-			    summary     = "Global request rate low",
-			    description = "The global request rate is low"
-			  }
-			  
-			foo = bar{label1="value1"}
-
-			ALERT BazAlert IF foo > 10
-			  ANNOTATIONS {
-			    description = "BazAlert",
-			    runbook     = "http://my.url",
-			    summary     = "Baz",
-			  }
-		`,
-		expected: Statements{
-			&RecordStmt{
-				Name: "dc:http_request:rate5m",
-				Expr: &AggregateExpr{
-					Op:       itemSum,
-					Grouping: []string{"dc"},
-					Expr: &Call{
-						Func: mustGetFunction("rate"),
-						Args: Expressions{
-							&MatrixSelector{
-								Name: "http_request_count",
-								LabelMatchers: []*labels.Matcher{
-									mustLabelMatcher(labels.MatchEqual, string(model.MetricNameLabel), "http_request_count"),
-								},
-								Range: 5 * time.Minute,
-							},
-						},
-					},
-				},
-				Labels: nil,
-			},
-			&AlertStmt{
-				Name: "GlobalRequestRateLow",
-				Expr: &ParenExpr{&BinaryExpr{
-					Op: itemLSS,
-					LHS: &VectorSelector{
-						Name: "dc:http_request:rate5m",
-						LabelMatchers: []*labels.Matcher{
-							mustLabelMatcher(labels.MatchEqual, string(model.MetricNameLabel), "dc:http_request:rate5m"),
-						},
-					},
-					RHS: &NumberLiteral{10000},
-				}},
-				Labels:   labels.FromStrings("service", "testservice"),
-				Duration: 5 * time.Minute,
-				Annotations: labels.FromStrings(
-					"summary", "Global request rate low",
-					"description", "The global request rate is low",
-				),
-			},
-			&RecordStmt{
-				Name: "foo",
-				Expr: &VectorSelector{
-					Name: "bar",
-					LabelMatchers: []*labels.Matcher{
-						mustLabelMatcher(labels.MatchEqual, "label1", "value1"),
-						mustLabelMatcher(labels.MatchEqual, string(model.MetricNameLabel), "bar"),
-					},
-				},
-			},
-			&AlertStmt{
-				Name: "BazAlert",
-				Expr: &BinaryExpr{
-					Op: itemGTR,
-					LHS: &VectorSelector{
-						Name: "foo",
-						LabelMatchers: []*labels.Matcher{
-							mustLabelMatcher(labels.MatchEqual, string(model.MetricNameLabel), "foo"),
-						},
-					},
-					RHS: &NumberLiteral{10},
-				},
-				Annotations: labels.FromStrings(
-					"summary", "Baz",
-					"description", "BazAlert",
-					"runbook", "http://my.url",
-				),
-			},
-		},
-	}, {
-		input: `foo{x="", a="z"} = bar{a="b", x=~"y"}`,
-		expected: Statements{
-			&RecordStmt{
-				Name: "foo",
-				Expr: &VectorSelector{
-					Name: "bar",
-					LabelMatchers: []*labels.Matcher{
-						mustLabelMatcher(labels.MatchEqual, "a", "b"),
-						mustLabelMatcher(labels.MatchRegexp, "x", "y"),
-						mustLabelMatcher(labels.MatchEqual, string(model.MetricNameLabel), "bar"),
-					},
-				},
-				Labels: labels.FromStrings("x", "", "a", "z"),
-			},
-		},
-	}, {
-		input: `ALERT SomeName IF some_metric > 1
-			LABELS {}
-			ANNOTATIONS {
-				summary = "Global request rate low",
-				description = "The global request rate is low",
-			}
-		`,
-		expected: Statements{
-			&AlertStmt{
-				Name: "SomeName",
-				Expr: &BinaryExpr{
-					Op: itemGTR,
-					LHS: &VectorSelector{
-						Name: "some_metric",
-						LabelMatchers: []*labels.Matcher{
-							mustLabelMatcher(labels.MatchEqual, string(model.MetricNameLabel), "some_metric"),
-						},
-					},
-					RHS: &NumberLiteral{1},
-				},
-				Labels: labels.Labels{},
-				Annotations: labels.FromStrings(
-					"summary", "Global request rate low",
-					"description", "The global request rate is low",
-				),
-			},
-		},
-	}, {
-		input: `
-			# A simple test alerting rule.
-			ALERT GlobalRequestRateLow IF(dc:http_request:rate5m < 10000) FOR 5
-			  LABELS {
-			    service = "testservice"
-			    # ... more fields here ...
-			  }
-			  ANNOTATIONS {
-			    summary = "Global request rate low"
-			    description = "The global request rate is low"
-			  }
-	  	`,
-		fail: true,
-	}, {
-		input:    "",
-		expected: Statements{},
-	}, {
-		input: "foo = time()",
-		expected: Statements{
-			&RecordStmt{
-				Name:   "foo",
-				Expr:   &Call{Func: mustGetFunction("time")},
-				Labels: nil,
-			}},
-	}, {
-		input: "foo = 1",
-		expected: Statements{
-			&RecordStmt{
-				Name:   "foo",
-				Expr:   &NumberLiteral{1},
-				Labels: nil,
-			}},
-	}, {
-		input: "foo = bar[5m]",
-		fail:  true,
-	}, {
-		input: `foo = "test"`,
-		fail:  true,
-	}, {
-		input: `foo = `,
-		fail:  true,
-	}, {
-		input: `foo{a!="b"} = bar`,
-		fail:  true,
-	}, {
-		input: `foo{a=~"b"} = bar`,
-		fail:  true,
-	}, {
-		input: `foo{a!~"b"} = bar`,
-		fail:  true,
-	},
-	// Fuzzing regression tests.
-	{
-		input: `I=-/`,
-		fail:  true,
-	},
-	{
-		input: `I=3E8/-=`,
-		fail:  true,
-	},
-	{
-		input: `M=-=-0-0`,
-		fail:  true,
-	},
-}
-
-func TestParseStatements(t *testing.T) {
-	for _, test := range testStatement {
-		parser := newParser(test.input)
-
-		stmts, err := parser.parseStmts()
-
-		// Unexpected errors are always caused by a bug.
-		if err == errUnexpected {
-			t.Fatalf("unexpected error occurred")
-		}
-
-		if !test.fail && err != nil {
-			t.Errorf("error in input: \n\n%s\n", test.input)
-			t.Fatalf("could not parse: %s", err)
-		}
-		if test.fail && err != nil {
-			continue
-		}
-
-		err = parser.typecheck(stmts)
-		if !test.fail && err != nil {
-			t.Errorf("error in input: \n\n%s\n", test.input)
-			t.Fatalf("typecheck failed: %s", err)
-		}
-
-		if test.fail {
-			if err != nil {
-				continue
-			}
-			t.Errorf("error in input: \n\n%s\n", test.input)
-			t.Fatalf("failure expected, but passed")
-		}
-
-		if !reflect.DeepEqual(stmts, test.expected) {
-			t.Errorf("error in input: \n\n%s\n", test.input)
-			t.Fatalf("no match\n\nexpected:\n%s\ngot: \n%s\n", Tree(test.expected), Tree(stmts))
-		}
-	}
-}
-
 func mustLabelMatcher(mt labels.MatchType, name, val string) *labels.Matcher {
 	m, err := labels.NewMatcher(mt, name, val)
 	if err != nil {
@@ -1772,6 +1690,39 @@ var testSeries = []struct {
 	}, {
 		input: `my_metric{a="b"} 1 3 _ 5 _a4`,
 		fail:  true,
+	}, {
+		input:          `my_metric{a="b"} 1 -1`,
+		expectedMetric: labels.FromStrings(labels.MetricName, "my_metric", "a", "b"),
+		expectedValues: newSeq(1, -1),
+	}, {
+		input:          `my_metric{a="b"} 1 +1`,
+		expectedMetric: labels.FromStrings(labels.MetricName, "my_metric", "a", "b"),
+		expectedValues: newSeq(1, 1),
+	}, {
+		input:          `my_metric{a="b"} 1 -1 -3-10x4 7 9 +5`,
+		expectedMetric: labels.FromStrings(labels.MetricName, "my_metric", "a", "b"),
+		expectedValues: newSeq(1, -1, -3, -13, -23, -33, -43, 7, 9, 5),
+	}, {
+		input:          `my_metric{a="b"} 1 +1 +4 -6 -2 8`,
+		expectedMetric: labels.FromStrings(labels.MetricName, "my_metric", "a", "b"),
+		expectedValues: newSeq(1, 1, 4, -6, -2, 8),
+	}, {
+		// Trailing spaces should be correctly handles.
+		input:          `my_metric{a="b"} 1 2 3    `,
+		expectedMetric: labels.FromStrings(labels.MetricName, "my_metric", "a", "b"),
+		expectedValues: newSeq(1, 2, 3),
+	}, {
+		input: `my_metric{a="b"} -3-3 -3`,
+		fail:  true,
+	}, {
+		input: `my_metric{a="b"} -3 -3-3`,
+		fail:  true,
+	}, {
+		input: `my_metric{a="b"} -3 _-2`,
+		fail:  true,
+	}, {
+		input: `my_metric{a="b"} -3 3+3x4-4`,
+		fail:  true,
 	},
 }
 
@@ -1791,30 +1742,24 @@ func newSeq(vals ...float64) (res []sequenceValue) {
 
 func TestParseSeries(t *testing.T) {
 	for _, test := range testSeries {
-		parser := newParser(test.input)
-		parser.lex.seriesDesc = true
-
-		metric, vals, err := parser.parseSeriesDesc()
+		metric, vals, err := parseSeriesDesc(test.input)
 
 		// Unexpected errors are always caused by a bug.
 		if err == errUnexpected {
 			t.Fatalf("unexpected error occurred")
 		}
 
-		if !test.fail && err != nil {
-			t.Errorf("error in input: \n\n%s\n", test.input)
-			t.Fatalf("could not parse: %s", err)
-		}
-		if test.fail && err != nil {
-			continue
-		}
-
 		if test.fail {
 			if err != nil {
 				continue
 			}
 			t.Errorf("error in input: \n\n%s\n", test.input)
 			t.Fatalf("failure expected, but passed")
+		} else {
+			if err != nil {
+				t.Errorf("error in input: \n\n%s\n", test.input)
+				t.Fatalf("could not parse: %s", err)
+			}
 		}
 
 		require.Equal(t, test.expectedMetric, metric)
@@ -1828,21 +1773,26 @@ func TestParseSeries(t *testing.T) {
 }
 
 func TestRecoverParserRuntime(t *testing.T) {
-	var p *parser
+	p := newParser("foo bar")
 	var err error
-	defer p.recover(&err)
 
+	defer func() {
+		if err != errUnexpected {
+			t.Fatalf("wrong error message: %q, expected %q", err, errUnexpected)
+		}
+
+		if _, ok := <-p.lex.items; ok {
+			t.Fatalf("lex.items was not closed")
+		}
+	}()
+	defer p.recover(&err)
 	// Cause a runtime panic.
 	var a []int
 	a[123] = 1
-
-	if err != errUnexpected {
-		t.Fatalf("wrong error message: %q, expected %q", err, errUnexpected)
-	}
 }
 
 func TestRecoverParserError(t *testing.T) {
-	var p *parser
+	p := newParser("foo bar")
 	var err error
 
 	e := fmt.Errorf("custom error")
