@@ -29,6 +29,8 @@ import (
 	"testing"
 	"time"
 
+	"github.com/prometheus/prometheus/pkg/relabel"
+
 	"github.com/prometheus/common/model"
 	"github.com/stretchr/testify/require"
 
@@ -37,6 +39,7 @@ import (
 	"github.com/prometheus/prometheus/config"
 	"github.com/prometheus/prometheus/discovery/targetgroup"
 	"github.com/prometheus/prometheus/pkg/labels"
+	"github.com/prometheus/prometheus/pkg/textparse"
 	"github.com/prometheus/prometheus/pkg/timestamp"
 	"github.com/prometheus/prometheus/pkg/value"
 	"github.com/prometheus/prometheus/storage"
@@ -67,10 +70,10 @@ func TestDroppedTargetsList(t *testing.T) {
 		cfg = &config.ScrapeConfig{
 			JobName:        "dropMe",
 			ScrapeInterval: model.Duration(1),
-			RelabelConfigs: []*config.RelabelConfig{
+			RelabelConfigs: []*relabel.Config{
 				{
-					Action:       config.RelabelDrop,
-					Regex:        mustNewRegexp("dropMe"),
+					Action:       relabel.Drop,
+					Regex:        relabel.MustNewRegexp("dropMe"),
 					SourceLabels: model.LabelNames{"job"},
 				},
 			},
@@ -78,7 +81,7 @@ func TestDroppedTargetsList(t *testing.T) {
 		tgs = []*targetgroup.Group{
 			{
 				Targets: []model.LabelSet{
-					model.LabelSet{model.AddressLabel: "127.0.0.1:9090"},
+					{model.AddressLabel: "127.0.0.1:9090"},
 				},
 			},
 		}
@@ -106,7 +109,7 @@ func TestDiscoveredLabelsUpdate(t *testing.T) {
 		ScrapeInterval: model.Duration(1),
 		ScrapeTimeout:  model.Duration(1),
 	}
-	sp.targets = make(map[uint64]*Target)
+	sp.activeTargets = make(map[uint64]*Target)
 	t1 := &Target{
 		discoveredLabels: labels.Labels{
 			labels.Label{
@@ -115,7 +118,7 @@ func TestDiscoveredLabelsUpdate(t *testing.T) {
 			},
 		},
 	}
-	sp.targets[t1.hash()] = t1
+	sp.activeTargets[t1.hash()] = t1
 
 	t2 := &Target{
 		discoveredLabels: labels.Labels{
@@ -127,7 +130,7 @@ func TestDiscoveredLabelsUpdate(t *testing.T) {
 	}
 	sp.sync([]*Target{t2})
 
-	testutil.Equals(t, t2.DiscoveredLabels(), sp.targets[t1.hash()].DiscoveredLabels())
+	testutil.Equals(t, t2.DiscoveredLabels(), sp.activeTargets[t1.hash()].DiscoveredLabels())
 }
 
 type testLoop struct {
@@ -145,9 +148,9 @@ func (l *testLoop) stop() {
 
 func TestScrapePoolStop(t *testing.T) {
 	sp := &scrapePool{
-		targets: map[uint64]*Target{},
-		loops:   map[uint64]loop{},
-		cancel:  func() {},
+		activeTargets: map[uint64]*Target{},
+		loops:         map[uint64]loop{},
+		cancel:        func() {},
 	}
 	var mtx sync.Mutex
 	stopped := map[uint64]bool{}
@@ -170,7 +173,7 @@ func TestScrapePoolStop(t *testing.T) {
 			mtx.Unlock()
 		}
 
-		sp.targets[t.hash()] = t
+		sp.activeTargets[t.hash()] = t
 		sp.loops[t.hash()] = l
 	}
 
@@ -198,8 +201,8 @@ func TestScrapePoolStop(t *testing.T) {
 	}
 	mtx.Unlock()
 
-	if len(sp.targets) > 0 {
-		t.Fatalf("Targets were not cleared on stopping: %d left", len(sp.targets))
+	if len(sp.activeTargets) > 0 {
+		t.Fatalf("Targets were not cleared on stopping: %d left", len(sp.activeTargets))
 	}
 	if len(sp.loops) > 0 {
 		t.Fatalf("Loops were not cleared on stopping: %d left", len(sp.loops))
@@ -218,7 +221,7 @@ func TestScrapePoolReload(t *testing.T) {
 	}
 	// On starting to run, new loops created on reload check whether their preceding
 	// equivalents have been stopped.
-	newLoop := func(_ *Target, s scraper) loop {
+	newLoop := func(_ *Target, s scraper, _ int, _ bool, _ []*relabel.Config) loop {
 		l := &testLoop{}
 		l.startFunc = func(interval, timeout time.Duration, errc chan<- error) {
 			if interval != 3*time.Second {
@@ -236,11 +239,11 @@ func TestScrapePoolReload(t *testing.T) {
 		return l
 	}
 	sp := &scrapePool{
-		appendable: &nopAppendable{},
-		targets:    map[uint64]*Target{},
-		loops:      map[uint64]loop{},
-		newLoop:    newLoop,
-		logger:     nil,
+		appendable:    &nopAppendable{},
+		activeTargets: map[uint64]*Target{},
+		loops:         map[uint64]loop{},
+		newLoop:       newLoop,
+		logger:        nil,
 	}
 
 	// Reloading a scrape pool with a new scrape configuration must stop all scrape
@@ -260,13 +263,13 @@ func TestScrapePoolReload(t *testing.T) {
 			mtx.Unlock()
 		}
 
-		sp.targets[t.hash()] = t
+		sp.activeTargets[t.hash()] = t
 		sp.loops[t.hash()] = l
 	}
 	done := make(chan struct{})
 
 	beforeTargets := map[uint64]*Target{}
-	for h, t := range sp.targets {
+	for h, t := range sp.activeTargets {
 		beforeTargets[h] = t
 	}
 
@@ -293,7 +296,7 @@ func TestScrapePoolReload(t *testing.T) {
 	}
 	mtx.Unlock()
 
-	if !reflect.DeepEqual(sp.targets, beforeTargets) {
+	if !reflect.DeepEqual(sp.activeTargets, beforeTargets) {
 		t.Fatalf("Reloading affected target states unexpectedly")
 	}
 	if len(sp.loops) != numTargets {
@@ -306,7 +309,12 @@ func TestScrapePoolAppender(t *testing.T) {
 	app := &nopAppendable{}
 	sp := newScrapePool(cfg, app, nil)
 
-	wrapped := sp.appender()
+	loop := sp.newLoop(&Target{}, nil, 0, false, nil)
+	appl, ok := loop.(*scrapeLoop)
+	if !ok {
+		t.Fatalf("Expected scrapeLoop but got %T", loop)
+	}
+	wrapped := appl.appender()
 
 	tl, ok := wrapped.(*timeLimitAppender)
 	if !ok {
@@ -316,9 +324,12 @@ func TestScrapePoolAppender(t *testing.T) {
 		t.Fatalf("Expected base appender but got %T", tl.Appender)
 	}
 
-	cfg.SampleLimit = 100
-
-	wrapped = sp.appender()
+	loop = sp.newLoop(&Target{}, nil, 100, false, nil)
+	appl, ok = loop.(*scrapeLoop)
+	if !ok {
+		t.Fatalf("Expected scrapeLoop but got %T", loop)
+	}
+	wrapped = appl.appender()
 
 	sl, ok := wrapped.(*limitAppender)
 	if !ok {
@@ -333,6 +344,46 @@ func TestScrapePoolAppender(t *testing.T) {
 	}
 }
 
+func TestScrapePoolRaces(t *testing.T) {
+	interval, _ := model.ParseDuration("500ms")
+	timeout, _ := model.ParseDuration("1s")
+	newConfig := func() *config.ScrapeConfig {
+		return &config.ScrapeConfig{ScrapeInterval: interval, ScrapeTimeout: timeout}
+	}
+	sp := newScrapePool(newConfig(), &nopAppendable{}, nil)
+	tgts := []*targetgroup.Group{
+		{
+			Targets: []model.LabelSet{
+				{model.AddressLabel: "127.0.0.1:9090"},
+				{model.AddressLabel: "127.0.0.2:9090"},
+				{model.AddressLabel: "127.0.0.3:9090"},
+				{model.AddressLabel: "127.0.0.4:9090"},
+				{model.AddressLabel: "127.0.0.5:9090"},
+				{model.AddressLabel: "127.0.0.6:9090"},
+				{model.AddressLabel: "127.0.0.7:9090"},
+				{model.AddressLabel: "127.0.0.8:9090"},
+			},
+		},
+	}
+
+	sp.Sync(tgts)
+	active := sp.ActiveTargets()
+	dropped := sp.DroppedTargets()
+	expectedActive, expectedDropped := len(tgts[0].Targets), 0
+	if len(sp.ActiveTargets()) != expectedActive {
+		t.Fatalf("Invalid number of active targets: expected %v, got %v", expectedActive, len(active))
+	}
+	if len(dropped) != expectedDropped {
+		t.Fatalf("Invalid number of dropped targets: expected %v, got %v", expectedDropped, len(dropped))
+	}
+
+	for i := 0; i < 20; i++ {
+		time.Sleep(time.Duration(10 * time.Millisecond))
+		sp.reload(newConfig())
+	}
+	sp.stop()
+}
+
 func TestScrapeLoopStopBeforeRun(t *testing.T) {
 	scraper := &testScraper{}
 
@@ -341,7 +392,7 @@ func TestScrapeLoopStopBeforeRun(t *testing.T) {
 		nil, nil,
 		nopMutator,
 		nopMutator,
-		nil,
+		nil, nil,
 	)
 
 	// The scrape pool synchronizes on stopping scrape loops. However, new scrape
@@ -404,6 +455,7 @@ func TestScrapeLoopStop(t *testing.T) {
 		nopMutator,
 		nopMutator,
 		app,
+		nil,
 	)
 
 	// Terminate loop after 2 scrapes.
@@ -434,7 +486,7 @@ func TestScrapeLoopStop(t *testing.T) {
 	if len(appender.result) < 5*3 || len(appender.result)%5 != 0 {
 		t.Fatalf("Expected at least 3 scrapes with 4 samples each, got %d samples", len(appender.result))
 	}
-	// All samples in a scrape must have the same timestmap.
+	// All samples in a scrape must have the same timestamp.
 	var ts int64
 	for i, s := range appender.result {
 		if i%5 == 0 {
@@ -468,6 +520,7 @@ func TestScrapeLoopRun(t *testing.T) {
 		nopMutator,
 		nopMutator,
 		app,
+		nil,
 	)
 
 	// The loop must terminate during the initial offset if the context
@@ -512,6 +565,7 @@ func TestScrapeLoopRun(t *testing.T) {
 		nopMutator,
 		nopMutator,
 		app,
+		nil,
 	)
 
 	go func() {
@@ -544,6 +598,54 @@ func TestScrapeLoopRun(t *testing.T) {
 	}
 }
 
+func TestScrapeLoopMetadata(t *testing.T) {
+	var (
+		signal  = make(chan struct{})
+		scraper = &testScraper{}
+		cache   = newScrapeCache()
+	)
+	defer close(signal)
+
+	ctx, cancel := context.WithCancel(context.Background())
+	sl := newScrapeLoop(ctx,
+		scraper,
+		nil, nil,
+		nopMutator,
+		nopMutator,
+		func() storage.Appender { return nopAppender{} },
+		cache,
+	)
+	defer cancel()
+
+	total, _, err := sl.append([]byte(`# TYPE test_metric counter
+# HELP test_metric some help text
+# UNIT test_metric metric
+test_metric 1
+# TYPE test_metric_no_help gauge
+# HELP test_metric_no_type other help text
+# EOF`), "application/openmetrics-text", time.Now())
+	testutil.Ok(t, err)
+	testutil.Equals(t, 1, total)
+
+	md, ok := cache.getMetadata("test_metric")
+	testutil.Assert(t, ok, "expected metadata to be present")
+	testutil.Assert(t, textparse.MetricTypeCounter == md.Type, "unexpected metric type")
+	testutil.Equals(t, "some help text", md.Help)
+	testutil.Equals(t, "metric", md.Unit)
+
+	md, ok = cache.getMetadata("test_metric_no_help")
+	testutil.Assert(t, ok, "expected metadata to be present")
+	testutil.Assert(t, textparse.MetricTypeGauge == md.Type, "unexpected metric type")
+	testutil.Equals(t, "", md.Help)
+	testutil.Equals(t, "", md.Unit)
+
+	md, ok = cache.getMetadata("test_metric_no_type")
+	testutil.Assert(t, ok, "expected metadata to be present")
+	testutil.Assert(t, textparse.MetricTypeUnknown == md.Type, "unexpected metric type")
+	testutil.Equals(t, "other help text", md.Help)
+	testutil.Equals(t, "", md.Unit)
+}
+
 func TestScrapeLoopRunCreatesStaleMarkersOnFailedScrape(t *testing.T) {
 	appender := &collectResultAppender{}
 	var (
@@ -560,6 +662,7 @@ func TestScrapeLoopRunCreatesStaleMarkersOnFailedScrape(t *testing.T) {
 		nopMutator,
 		nopMutator,
 		app,
+		nil,
 	)
 	// Succeed once, several failures, then stop.
 	numScrapes := 0
@@ -617,6 +720,7 @@ func TestScrapeLoopRunCreatesStaleMarkersOnParseFailure(t *testing.T) {
 		nopMutator,
 		nopMutator,
 		app,
+		nil,
 	)
 
 	// Succeed once, several failures, then stop.
@@ -706,11 +810,6 @@ func TestScrapeLoopAppend(t *testing.T) {
 
 	for _, test := range tests {
 		app := &collectResultAppender{}
-		sp := &scrapePool{
-			config: &config.ScrapeConfig{
-				HonorLabels: test.honorLabels,
-			},
-		}
 
 		discoveryLabels := &Target{
 			labels: labels.FromStrings(test.discoveryLabels...),
@@ -719,17 +818,18 @@ func TestScrapeLoopAppend(t *testing.T) {
 		sl := newScrapeLoop(context.Background(),
 			nil, nil, nil,
 			func(l labels.Labels) labels.Labels {
-				return sp.mutateSampleLabels(l, discoveryLabels)
+				return mutateSampleLabels(l, discoveryLabels, test.honorLabels, nil)
 			},
 			func(l labels.Labels) labels.Labels {
-				return sp.mutateReportSampleLabels(l, discoveryLabels)
+				return mutateReportSampleLabels(l, discoveryLabels)
 			},
 			func() storage.Appender { return app },
+			nil,
 		)
 
 		now := time.Now()
 
-		_, _, err := sl.append([]byte(test.scrapeLabels), now)
+		_, _, err := sl.append([]byte(test.scrapeLabels), "", now)
 		if err != nil {
 			t.Fatalf("Unexpected append error: %s", err)
 		}
@@ -763,6 +863,7 @@ func TestScrapeLoopAppendSampleLimit(t *testing.T) {
 		nopMutator,
 		nopMutator,
 		func() storage.Appender { return app },
+		nil,
 	)
 
 	// Get the value of the Counter before performing the append.
@@ -774,7 +875,7 @@ func TestScrapeLoopAppendSampleLimit(t *testing.T) {
 	beforeMetricValue := beforeMetric.GetCounter().GetValue()
 
 	now := time.Now()
-	_, _, err = sl.append([]byte("metric_a 1\nmetric_b 1\nmetric_c 1\n"), now)
+	_, _, err = sl.append([]byte("metric_a 1\nmetric_b 1\nmetric_c 1\n"), "", now)
 	if err != errSampleLimit {
 		t.Fatalf("Did not see expected sample limit error: %s", err)
 	}
@@ -822,14 +923,15 @@ func TestScrapeLoop_ChangingMetricString(t *testing.T) {
 		nopMutator,
 		nopMutator,
 		func() storage.Appender { return capp },
+		nil,
 	)
 
 	now := time.Now()
-	_, _, err = sl.append([]byte(`metric_a{a="1",b="1"} 1`), now)
+	_, _, err = sl.append([]byte(`metric_a{a="1",b="1"} 1`), "", now)
 	if err != nil {
 		t.Fatalf("Unexpected append error: %s", err)
 	}
-	_, _, err = sl.append([]byte(`metric_a{b="1",a="1"} 2`), now.Add(time.Minute))
+	_, _, err = sl.append([]byte(`metric_a{b="1",a="1"} 2`), "", now.Add(time.Minute))
 	if err != nil {
 		t.Fatalf("Unexpected append error: %s", err)
 	}
@@ -860,14 +962,15 @@ func TestScrapeLoopAppendStaleness(t *testing.T) {
 		nopMutator,
 		nopMutator,
 		func() storage.Appender { return app },
+		nil,
 	)
 
 	now := time.Now()
-	_, _, err := sl.append([]byte("metric_a 1\n"), now)
+	_, _, err := sl.append([]byte("metric_a 1\n"), "", now)
 	if err != nil {
 		t.Fatalf("Unexpected append error: %s", err)
 	}
-	_, _, err = sl.append([]byte(""), now.Add(time.Second))
+	_, _, err = sl.append([]byte(""), "", now.Add(time.Second))
 	if err != nil {
 		t.Fatalf("Unexpected append error: %s", err)
 	}
@@ -904,14 +1007,15 @@ func TestScrapeLoopAppendNoStalenessIfTimestamp(t *testing.T) {
 		nopMutator,
 		nopMutator,
 		func() storage.Appender { return app },
+		nil,
 	)
 
 	now := time.Now()
-	_, _, err := sl.append([]byte("metric_a 1 1000\n"), now)
+	_, _, err := sl.append([]byte("metric_a 1 1000\n"), "", now)
 	if err != nil {
 		t.Fatalf("Unexpected append error: %s", err)
 	}
-	_, _, err = sl.append([]byte(""), now.Add(time.Second))
+	_, _, err = sl.append([]byte(""), "", now.Add(time.Second))
 	if err != nil {
 		t.Fatalf("Unexpected append error: %s", err)
 	}
@@ -942,6 +1046,7 @@ func TestScrapeLoopRunReportsTargetDownOnScrapeError(t *testing.T) {
 		nopMutator,
 		nopMutator,
 		app,
+		nil,
 	)
 
 	scraper.scrapeFunc = func(ctx context.Context, w io.Writer) error {
@@ -970,6 +1075,7 @@ func TestScrapeLoopRunReportsTargetDownOnInvalidUTF8(t *testing.T) {
 		nopMutator,
 		nopMutator,
 		app,
+		nil,
 	)
 
 	scraper.scrapeFunc = func(ctx context.Context, w io.Writer) error {
@@ -1015,10 +1121,11 @@ func TestScrapeLoopAppendGracefullyIfAmendOrOutOfOrderOrOutOfBounds(t *testing.T
 		nopMutator,
 		nopMutator,
 		func() storage.Appender { return app },
+		nil,
 	)
 
 	now := time.Unix(1, 0)
-	_, _, err := sl.append([]byte("out_of_order 1\namend 1\nnormal 1\nout_of_bounds 1\n"), now)
+	_, _, err := sl.append([]byte("out_of_order 1\namend 1\nnormal 1\nout_of_bounds 1\n"), "", now)
 	if err != nil {
 		t.Fatalf("Unexpected append error: %s", err)
 	}
@@ -1047,10 +1154,11 @@ func TestScrapeLoopOutOfBoundsTimeError(t *testing.T) {
 				maxTime:  timestamp.FromTime(time.Now().Add(10 * time.Minute)),
 			}
 		},
+		nil,
 	)
 
 	now := time.Now().Add(20 * time.Minute)
-	total, added, err := sl.append([]byte("normal 1\n"), now)
+	total, added, err := sl.append([]byte("normal 1\n"), "", now)
 	if total != 1 {
 		t.Error("expected 1 metric")
 		return
@@ -1074,8 +1182,8 @@ func TestTargetScraperScrapeOK(t *testing.T) {
 	server := httptest.NewServer(
 		http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
 			accept := r.Header.Get("Accept")
-			if !strings.HasPrefix(accept, "text/plain;") {
-				t.Errorf("Expected Accept header to prefer text/plain, got %q", accept)
+			if !strings.HasPrefix(accept, "application/openmetrics-text;") {
+				t.Errorf("Expected Accept header to prefer application/openmetrics-text, got %q", accept)
 			}
 
 			timeout := r.Header.Get("X-Prometheus-Scrape-Timeout-Seconds")
@@ -1106,9 +1214,11 @@ func TestTargetScraperScrapeOK(t *testing.T) {
 	}
 	var buf bytes.Buffer
 
-	if err := ts.scrape(context.Background(), &buf); err != nil {
+	contentType, err := ts.scrape(context.Background(), &buf)
+	if err != nil {
 		t.Fatalf("Unexpected scrape error: %s", err)
 	}
+	require.Equal(t, "text/plain; version=0.0.4", contentType)
 	require.Equal(t, "metric_a 1\nmetric_b 2\n", buf.String())
 }
 
@@ -1146,8 +1256,11 @@ func TestTargetScrapeScrapeCancel(t *testing.T) {
 	}()
 
 	go func() {
-		if err := ts.scrape(ctx, ioutil.Discard); err != context.Canceled {
-			errc <- fmt.Errorf("Expected context cancelation error but got: %s", err)
+		_, err := ts.scrape(ctx, ioutil.Discard)
+		if err == nil {
+			errc <- fmt.Errorf("Expected error but got nil")
+		} else if ctx.Err() != context.Canceled {
+			errc <- fmt.Errorf("Expected context cancelation error but got: %s", ctx.Err())
 		}
 		close(errc)
 	}()
@@ -1161,7 +1274,7 @@ func TestTargetScrapeScrapeCancel(t *testing.T) {
 		}
 	}
 	// If this is closed in a defer above the function the test server
-	// does not terminate and the test doens't complete.
+	// doesn't terminate and the test doesn't complete.
 	close(block)
 }
 
@@ -1188,7 +1301,7 @@ func TestTargetScrapeScrapeNotFound(t *testing.T) {
 		client: http.DefaultClient,
 	}
 
-	if err := ts.scrape(context.Background(), ioutil.Discard); !strings.Contains(err.Error(), "404") {
+	if _, err := ts.scrape(context.Background(), ioutil.Discard); !strings.Contains(err.Error(), "404") {
 		t.Fatalf("Expected \"404 NotFound\" error but got: %s", err)
 	}
 }
@@ -1216,9 +1329,9 @@ func (ts *testScraper) report(start time.Time, duration time.Duration, err error
 	ts.lastError = err
 }
 
-func (ts *testScraper) scrape(ctx context.Context, w io.Writer) error {
+func (ts *testScraper) scrape(ctx context.Context, w io.Writer) (string, error) {
 	if ts.scrapeFunc != nil {
-		return ts.scrapeFunc(ctx, w)
+		return "", ts.scrapeFunc(ctx, w)
 	}
-	return ts.scrapeErr
+	return "", ts.scrapeErr
 }
