@@ -15,6 +15,7 @@ package storage
 
 import (
 	"fmt"
+	"math"
 	"testing"
 
 	"github.com/stretchr/testify/require"
@@ -97,8 +98,18 @@ func TestMergeSeriesSet(t *testing.T) {
 				newMockSeries(labels.FromStrings("foo", "bar"), []sample{{0, 0}, {1, 1}, {2, 2}, {3, 3}}),
 			),
 		},
+		{
+			input: []SeriesSet{newMockSeriesSet(
+				newMockSeries(labels.FromStrings("foo", "bar"), []sample{{0, math.NaN()}}),
+			), newMockSeriesSet(
+				newMockSeries(labels.FromStrings("foo", "bar"), []sample{{0, math.NaN()}}),
+			)},
+			expected: newMockSeriesSet(
+				newMockSeries(labels.FromStrings("foo", "bar"), []sample{{0, math.NaN()}}),
+			),
+		},
 	} {
-		merged := NewMergeSeriesSet(tc.input)
+		merged := NewMergeSeriesSet(tc.input, nil)
 		for merged.Next() {
 			require.True(t, tc.expected.Next())
 			actualSeries := merged.At()
@@ -136,6 +147,14 @@ func TestMergeIterator(t *testing.T) {
 			},
 			expected: []sample{{0, 0}, {1, 1}, {2, 2}, {3, 3}, {4, 4}, {5, 5}},
 		},
+		{
+			input: []SeriesIterator{
+				newListSeriesIterator([]sample{{0, 0}, {1, 1}}),
+				newListSeriesIterator([]sample{{0, 0}, {2, 2}}),
+				newListSeriesIterator([]sample{{2, 2}, {3, 3}}),
+			},
+			expected: []sample{{0, 0}, {1, 1}, {2, 2}, {3, 3}},
+		},
 	} {
 		merged := newMergeIterator(tc.input)
 		actual := drainSamples(merged)
@@ -189,6 +208,10 @@ func drainSamples(iter SeriesIterator) []sample {
 	result := []sample{}
 	for iter.Next() {
 		t, v := iter.At()
+		// NaNs can't be compared normally, so substitute for another value.
+		if math.IsNaN(v) {
+			v = -42
+		}
 		result = append(result, sample{t, v})
 	}
 	return result
@@ -239,7 +262,7 @@ func makeMergeSeriesSet(numSeriesSets, numSeries, numSamples int) SeriesSet {
 	for i := 0; i < numSeriesSets; i++ {
 		seriesSets = append(seriesSets, makeSeriesSet(numSeries, numSamples))
 	}
-	return NewMergeSeriesSet(seriesSets)
+	return NewMergeSeriesSet(seriesSets, nil)
 }
 
 func benchmarkDrain(seriesSet SeriesSet, b *testing.B) {
