@@ -19,7 +19,7 @@ import (
 	"github.com/prometheus/prometheus/storage"
 )
 
-// Appender implements retrieval.Appendable.
+// Appender implements scrape.Appendable.
 func (s *Storage) Appender() (storage.Appender, error) {
 	return s, nil
 }
@@ -29,11 +29,13 @@ func (s *Storage) Add(l labels.Labels, t int64, v float64) (uint64, error) {
 	s.mtx.RLock()
 	defer s.mtx.RUnlock()
 	for _, q := range s.queues {
-		q.Append(&model.Sample{
+		if err := q.Append(&model.Sample{
 			Metric:    labelsToMetric(l),
 			Timestamp: model.Time(t),
 			Value:     model.SampleValue(v),
-		})
+		}); err != nil {
+			panic(err) // QueueManager.Append() should always return nil as per doc string.
+		}
 	}
 	return 0, nil
 }
