@@ -18,8 +18,8 @@ import (
 	"fmt"
 	"math"
 	"math/rand"
+	"net"
 	"net/http"
-	"net/url"
 	"os"
 	"path/filepath"
 	"time"
@@ -36,41 +36,24 @@ import (
 	"github.com/prometheus/tsdb"
 	tsdbLabels "github.com/prometheus/tsdb/labels"
 
-	"github.com/prometheus/prometheus/pkg/labels"
 	"github.com/prometheus/prometheus/pkg/timestamp"
 	pb "github.com/prometheus/prometheus/prompb"
-	"github.com/prometheus/prometheus/promql"
-	"github.com/prometheus/prometheus/scrape"
-	"github.com/prometheus/prometheus/storage"
 )
 
 // API encapsulates all API services.
 type API struct {
-	enableAdmin   bool
-	now           func() time.Time
-	db            func() *tsdb.DB
-	q             func(ctx context.Context, mint, maxt int64) (storage.Querier, error)
-	targets       func() []*scrape.Target
-	alertmanagers func() []*url.URL
+	enableAdmin bool
+	db          func() *tsdb.DB
 }
 
 // New returns a new API object.
 func New(
-	now func() time.Time,
 	db func() *tsdb.DB,
-	qe *promql.Engine,
-	q func(ctx context.Context, mint, maxt int64) (storage.Querier, error),
-	targets func() []*scrape.Target,
-	alertmanagers func() []*url.URL,
 	enableAdmin bool,
 ) *API {
 	return &API{
-		now:           now,
-		db:            db,
-		q:             q,
-		targets:       targets,
-		alertmanagers: alertmanagers,
-		enableAdmin:   enableAdmin,
+		db:          db,
+		enableAdmin: enableAdmin,
 	}
 }
 
@@ -84,13 +67,17 @@ func (api *API) RegisterGRPC(srv *grpc.Server) {
 }
 
 // HTTPHandler returns an HTTP handler for a REST API gateway to the given grpc address.
-func (api *API) HTTPHandler(grpcAddr string) (http.Handler, error) {
-	ctx := context.Background()
-
+func (api *API) HTTPHandler(ctx context.Context, grpcAddr string) (http.Handler, error) {
 	enc := new(protoutil.JSONPb)
 	mux := runtime.NewServeMux(runtime.WithMarshalerOption(enc.ContentType(), enc))
 
-	opts := []grpc.DialOption{grpc.WithInsecure()}
+	opts := []grpc.DialOption{
+		grpc.WithInsecure(),
+		// Replace the default dialer that connects through proxy when HTTP_PROXY is set.
+		grpc.WithDialer(func(addr string, _ time.Duration) (net.Conn, error) {
+			return (&net.Dialer{}).DialContext(ctx, "tcp", addr)
+		}),
+	}
 
 	err := pb.RegisterAdminHandlerFromEndpoint(ctx, mux, grpcAddr, opts)
 	if err != nil {
@@ -124,18 +111,8 @@ var (
 	maxTime = time.Unix(math.MaxInt64/1000-62135596801, 999999999)
 )
 
-func labelsToProto(lset labels.Labels) pb.Labels {
-	r := pb.Labels{
-		Labels: make([]pb.Label, 0, len(lset)),
-	}
-	for _, l := range lset {
-		r.Labels = append(r.Labels, pb.Label{Name: l.Name, Value: l.Value})
-	}
-	return r
-}
-
 // AdminDisabled implements the administration interface that informs
-// that the API endpoints are disbaled.
+// that the API endpoints are disabled.
 type AdminDisabled struct {
 }
 
@@ -149,7 +126,7 @@ func (s *AdminDisabled) TSDBCleanTombstones(_ old_ctx.Context, _ *pb.TSDBCleanTo
 	return nil, status.Error(codes.Unavailable, "Admin APIs are disabled")
 }
 
-// DeleteSeries imeplements pb.AdminServer.
+// DeleteSeries implements pb.AdminServer.
 func (s *AdminDisabled) DeleteSeries(_ old_ctx.Context, r *pb.SeriesDeleteRequest) (*pb.SeriesDeleteResponse, error) {
 	return nil, status.Error(codes.Unavailable, "Admin APIs are disabled")
 }
